import unittest

from absynthe.cfg import LoggerNode, SimpleLoggerNode


class LoggerNodeTest(unittest.TestCase):

    def test_abstractClassError(self):
        kwargs = {LoggerNode.KW_CORE_CLASS_NAME: "UniformNode"}
        with self.assertRaises(TypeError):
            _ = LoggerNode("testLogger", **kwargs)
        return


class SimpleLoggerNodeTest(unittest.TestCase):

    def test_basicTest(self):
        kwargs = {LoggerNode.KW_CORE_CLASS_NAME: "UniformNode",
                  SimpleLoggerNode.KW_PREFIX: "method1()"}
        _ = SimpleLoggerNode("testLogger", **kwargs)
        # _.printDebugInfo()
        return

    def test_logGeneration(self):
        kwargs = {LoggerNode.KW_CORE_CLASS_NAME: "UniformNode"}
        testLogger = SimpleLoggerNode("testLogger", **kwargs)
        params: list = ["test_param ", "123 ", "more test ", "456.789"]

        # Assert that the same log messages is generated for identical arguments
        self.assertEqual(testLogger.logInfo("11-May-2019, 11:31:13:2323", params),
                         testLogger.logInfo("11-May-2019, 11:31:13:2323", params))

        # Assert that the same error messages is generated for identical arguments
        self.assertEqual(testLogger.logError("11-May-2019, 11:31:13:2323", params),
                         testLogger.logError("11-May-2019, 11:31:13:2323", params))
        return


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import

# from .graph_builder import GraphBuilder, TreeBuilder, DAGBuilder, DCGBuilder
# from .behavior import Behavior, MonospaceInterleaving

# __all__ = ["GraphBuilder", "TreeBuilder", "DAGBuilder", "DCGBuilder",
#            "Behavior", "MonospaceInterleaving"]



from __future__ import print_function
from sys import stderr
from typing import TextIO, List, Tuple

from .node import Node
from random import randint


class Graph(object):
    """
    The Graph class. Can have multiple root nodes; and it suffices
    for objects of this class to only keep track of the root nodes.

    The actual graph is defined by recursively following the
    successors of all the roots.

    NOTE: The nodes in the graph must all have unique IDs, otherwise
          it will result in unintended behaviour. For instance, the
          size() method will not work properly since it relies on
          unique IDs to count the number of nodes.
    """

    def __init__(self, id: str, numRoots: int = 1) -> None:
        """
        Constructor for the class.

        Args:
          id(str): A unique ID for the graph
          numRoots(int): The number of roots that this graph contains.
        """
        self._id = id
        self._numRoots = numRoots
        self._roots = []
        return

    def addRoot(self, root: Node):
        """
        Adds a root to this graph.
        Args:
          root(Node): A Node object to be added to the list of roots.
        Raises:
          IndexError: In case the list of roots already contains
                      _numRoots nodes in it.
        """
        if (len(self._roots) == self._numRoots):
            raise IndexError("Cannot accommodate any more root nodes.")

        self._roots.append(root)
        return

    def getRootAt(self, index: int = 0) -> Node:
        """
        Args:
          index(int): The index of the desired root, mindful of the order
                    in which roots were added to this graph.
        Returns:
          Node: The root at the given index.
        Raises:
          IndexError: If the index is out of bounds.
        """
        try:
            return self._roots[index]
        except IndexError as error:
            print("Illegal argument for index:", index,
                  "Number of root nodes:", len(self._roots),
                  "Expected number of root nodes:", self._numRoots,
                  file=stderr)
            raise error
        return None

    def getRootAtRandom(self) -> Node:
        """
        Returns:
          Node: One of the root nodes of this graph, selecte uniformly
                at random.
        """
        randomIndex = randint(0, len(self._roots) - 1)
        return self._roots[randomIndex]

    def getID(self) -> str:
        return self._id

    def _bfsAndCount(self, node: Node) -> int:
        """
        Search the tree located below 'node' in a breadth-first manner,
        and return the count of this sub-tree, incl. 'node'in the count.

        Args:
          node(Node): The node, the size of whose sub-tree is to be
                      computed.
        Returns:
          int: The number of nodes in the sub-tree rooted at 'node',
               including 'node'.
        """
        nodeID = node.getID()
        try:
            _ = self._nodeDict[nodeID]
            return 0
        except KeyError:
            self._nodeDict[nodeID] = True
            pass
        count = 1

        for i in range(node.getNumSuccessors()):
            count += self._bfsAndCount(node.getSuccessorAt(i))
        return count

    def size(self) -> int:
        """
        Returns:
          int: The total number of nodes (with unique IDs) in this graph.
               If there are multiple Node objects with identical IDs,
               then this method will not produce expected result.
        """
        size = 0
        self._nodeDict = dict()
        for r in self._roots:
            if r is not None:
                size += self._bfsAndCount(r)
        return size

    def _bfsAndAdd(self, node: Node, transitionList: List[Tuple[str, str]]) -> None:
        nodeID = node.getID()
        try:
            _ = self._nodeDict[nodeID]
            return
        except KeyError:
            self._nodeDict[nodeID] = True
            pass

        for i in range(node.getNumSuccessors()):
            succNode: Node = node.getSuccessorAt(i)
            succNodeID = succNode.getID()
            transitionList.append((nodeID, succNodeID))
            self._bfsAndAdd(succNode, transitionList)

        return

    def dumpDotFile(self, fp: TextIO) -> None:
        """
        Creates a file that could be visualised using graphviz's DOT program.
        Args:
          fp(typing.TextIO): A text file stream that can be written to.
        """
        self._nodeDict = dict()
        transitionList: List[Tuple[str, str]] = list()

        for r in self._roots:
            if r is not None:
                self._bfsAndAdd(r, transitionList)

        fileContent: List[str] = ["digraph \""]
        fileContent.append(self._id)
        fileContent.append("\" {\n")

        for transition in transitionList:
            fileContent.append("\t\"")
            fileContent.append(transition[0])
            fileContent.append("\" -> \"")
            fileContent.append(transition[1])
            fileContent.append("\";\n")

        fileContent.append("}\n")

        fileText = "".join(fileContent)
        fp.write(fileText)
        fp.flush()
        return



from __future__ import print_function
from typing import List
from random import randint
from datetime import datetime

from .cfg import LoggerNode, Graph

# Imports for Behavior
from abc import ABC, abstractmethod


class Behavior(ABC):

    @abstractmethod
    def addGraph(self, graph: Graph) -> None:
        pass

    @abstractmethod
    def synthesize(self, numRuns: int) -> None:
        pass


class MonospaceInterleaving(Behavior):

    def __init__(self, withSessions: bool = False):
        self._cfgList: List[Graph] = list()
        self._fixedTimeDelta: float = 0.05
        self._inclSessionID: bool = withSessions
        return

    def addGraph(self, graph: Graph) -> None:
        self._cfgList.append(graph)
        return

    def synthesize(self, numRuns: int = 100):
        numGraphs: int = len(self._cfgList)
        nextNodeOf: List[LoggerNode] = None
        graphIdx: int = -1
        wallClock: float = -2.5
        for i in range(numRuns):  # Complete a traversal of each graph
            nextNodeOf = [self._cfgList[i].getRootAtRandom() for i in range(numGraphs)]
            graphsAvailable = list(range(numGraphs))  # Shrinks as we reach the leaf of a graph
            toTraverse: int = numGraphs
            wallClock += 2.5  # Adding time delay between successive runs
            while 0 < toTraverse:
                # If there are still graphs available to traverse,
                # randomly choose a graph among those whose leaves
                # have not yet been reached in this run.
                posInGraphsAvailable: int = randint(0, toTraverse - 1)
                graphIdx: int = graphsAvailable[posInGraphsAvailable]
                graph: Graph = self._cfgList[graphIdx]

                timeStamp: str = str(datetime.fromtimestamp(wallClock))
                sessionID: str = ""
                if self._inclSessionID:
                    sessionID = "_".join(["SESSION", str(i), str(graphIdx)])

                # For the sake of better readability of logs, append
                # graph ID to the time stamp.
                logPrefix: str = " ".join([timeStamp, sessionID, graph.getID()])
                node: LoggerNode = nextNodeOf[graphIdx]
                yield node.logInfo(logPrefix, None)

                wallClock += self._fixedTimeDelta
                nextNode: LoggerNode = node.getSuccessorAtRandom()
                nextNodeOf[graphIdx] = nextNode
                if nextNode is None:
                    _ = graphsAvailable.pop(posInGraphsAvailable)
                    toTraverse -= 1



import unittest
import os

from absynthe.graph_builder import TreeBuilder, DCGBuilder
from absynthe.behavior import MonospaceInterleaving


class test_monospaceInterleaving(unittest.TestCase):

    def test_treeLogGeneration(self):
        tree_kwargs = {TreeBuilder.KW_NUM_ROOTS: "2",
                       TreeBuilder.KW_NUM_LEAVES: "4",
                       TreeBuilder.KW_BRANCHING_DEGREE: "2",
                       TreeBuilder.KW_NUM_INNER_NODES: "16",
                       TreeBuilder.KW_SUPPORTED_NODE_TYPES: "SimpleLoggerNode"}

        simpleTreeBuilder = TreeBuilder(**tree_kwargs)

        wSessionID: bool = True
        testBehavior = MonospaceInterleaving(wSessionID)
        testBehavior.addGraph(simpleTreeBuilder.generateNewGraph())
        testBehavior.addGraph(simpleTreeBuilder.generateNewGraph())
        testBehavior.addGraph(simpleTreeBuilder.generateNewGraph())
        testBehavior.addGraph(simpleTreeBuilder.generateNewGraph())

        fileName = os.path.dirname(os.getcwd() +
                                   "/test/resources/") + "/Tree_MonospaceInterleaving_test.log"
        with open(fileName, 'w') as logfile:
            for logLine in testBehavior.synthesize(2):
                logfile.write(logLine)
                logfile.write(os.linesep)

        return

    def test_DCGLogGeneration(self):
        dcg_kwargs = {DCGBuilder.KW_NUM_ROOTS: "2",
                      DCGBuilder.KW_NUM_LEAVES: "1",
                      DCGBuilder.KW_BRANCHING_DEGREE: "2",
                      DCGBuilder.KW_NUM_INNER_NODES: "16",
                      DCGBuilder.KW_SUPPORTED_NODE_TYPES: "SimpleLoggerNode"}

        simpleDCGBuilder = DCGBuilder(**dcg_kwargs)

        wSessionID: bool = True
        testBehavior = MonospaceInterleaving(wSessionID)
        testBehavior.addGraph(simpleDCGBuilder.generateNewGraph())
        testBehavior.addGraph(simpleDCGBuilder.generateNewGraph())
        testBehavior.addGraph(simpleDCGBuilder.generateNewGraph())
        testBehavior.addGraph(simpleDCGBuilder.generateNewGraph())

        fileName = os.path.dirname(os.getcwd() +
                                   "/test/resources/") + "/DCG_MonospaceInterleaving_test.log"
        with open(fileName, 'w') as logfile:
            for logLine in testBehavior.synthesize(2):
                logfile.write(logLine)
                logfile.write(os.linesep)

        return



from __future__ import print_function
from sys import stderr
from typing import List
from random import randint

# Local imports
from .cfg import Node, Graph
from .cfg.utils import Utils

# Imports for GraphBuilder
from abc import ABC, abstractmethod
from sys import modules
import inspect

# Imports for concrete graph builders
from collections import defaultdict
from random import sample
from math import ceil


class GraphBuilder(ABC):
    """
    The abstract base class for all kinds of graph builders. Each graph builder object takes
    keyword arguments to initialise itself. The base class only specifies one method,
    `generateNewGraph()`, which should be able to generate a *new* graph with every invocation
    (subject to the initialising parameters).

    This class can be extended in two ways:

    1. As illustrated by the TreeBuilder class and its subclasses, implementations that take
       a number of arguments and randomly create new graphs.
    2. TODO: Implementations that read hand-crafted specifications and creates graphs that
       meet those specs.
    """

    def __init__(self, **kwargs: str) -> None:
        pass

    @abstractmethod
    def generateNewGraph(self, **kwargs: str) -> Graph:
        """
        Generates a new graph adhering to the arguments specified by the kwargs
        in the constructor. Each invocation of this method can potentially
        generate a different graph.
        """
        pass


class TreeBuilder(GraphBuilder):

    numGraphs: int = 0
    numNodes: int = 0

    # Keywords for the kwargs expected by the constructor
    KW_NUM_ROOTS = "NUM_ROOTS"
    KW_NUM_LEAVES = "NUM_LEAVES"
    KW_NUM_INNER_NODES = "NUM_INNER_NODES"
    KW_BRANCHING_DEGREE = "BRANCHING_DEGREE"
    KW_SUPPORTED_NODE_TYPES = "SUPPORTED_NODE_TYPES"

    def __init__(self, **kwargs: str) -> None:
        """
        Construtor for the class.
        Args:
          **kwargs(str): Must have the following keywords.
            KW_NUM_ROOTS - specifying the number of roots in any graph this builder creates.
            KW_NUM_LEAVES - specifying the number of leaves that any graph can have.
            KW_NUM_INNER_NODES - Besides roots and leaves, the approx. number of inner nodes.
            KW_BRANCHING_DEGREE - Approximate branching factor for each node.
            KW_SUPPORTED_NODE_TYPES - Concrete LoggerNode types that can be part of graphs.
        """
        # Fields that would be initialised from user input.
        # Failure to initialise any one of these will raise an error.
        self._numRoots: int = 0
        self._numLeaves: int = 0
        self._numInnerNodes: int = 0
        self._branchingDegree: int = 0
        self._deltaRange: int = 0
        self._supportedNodeTypes: List[str] = None
        self._rootList: List[Node] = None

        className: str = type(self).__name__
        try:
            self._numRoots = int(kwargs[TreeBuilder.KW_NUM_ROOTS])
        except KeyError as ke:
            print(className,
                  "ERROR - Number of roots must be specified using the key",
                  className, "\b.KW_NUM_ROOTS.",
                  file=stderr)
            raise ke

        try:
            self._numLeaves = int(kwargs[TreeBuilder.KW_NUM_LEAVES])
        except KeyError as ke:
            print(className,
                  "ERROR - Number of leaves must be specified using the key",
                  className, "\b.KW_NUM_LEAVES.",
                  file=stderr)
            raise ke

        try:
            self._numInnerNodes = int(kwargs[TreeBuilder.KW_NUM_INNER_NODES])
        except KeyError as ke:
            print(className,
                  "ERROR - Number of inner nodes must be specified using the key",
                  className, "\b.KW_NUM_INNER_NODES.",
                  file=stderr)
            raise ke

        try:
            self._branchingDegree = int(kwargs[TreeBuilder.KW_BRANCHING_DEGREE])
            self._deltaRange = self._branchingDegree // 2
        except KeyError as ke:
            print(className,
                  "ERROR - Approx. branching degree must be specified using the key",
                  className, "\b.KW_BRANCHING_DEGREE.",
                  file=stderr)
            raise ke

        try:
            self._supportedNodeTypes = [x.strip()
                                        for x in
                                        kwargs[TreeBuilder.KW_SUPPORTED_NODE_TYPES].split(',')]
        except KeyError as ke:
            print(className,
                  "ERROR- Types of nodes supported in this graph must be specified using the key",
                  className, "\b.KW_SUPPORTED_NODE_TYPES as comma-separated list of class names.",
                  file=stderr)
            raise ke

        self._coreNodeClasses = list()
        for name, obj in inspect.getmembers(modules["absynthe.cfg.node"], inspect.isclass):
            if not inspect.isabstract(obj) and not name == "ABC":
                self._coreNodeClasses.append(name)

        # Everything okay, so instantiate object
        return

    def _howManySuccessors(self, succLimit: int = None) -> int:
        """
        Randomly generate the number indicating how many successors a node will have. The
        following possibilities are equally likely:
        1. One successor
        2. Two successors
        3. (BRANCHING_DEGREE +/- delta) successors
        Returns:
          int: The number of successors a node will have
        """
        if succLimit is not None:
            return randint(0, succLimit - 1)

        theseMany: int = randint(1, 3)
        if 3 > theseMany:
            return theseMany

        delta: int = randint(-self._deltaRange, self._deltaRange)
        theseMany = self._branchingDegree + delta
        return theseMany

    def _makeConnections(self, fromLayer: List[Node], toLayer: List[Node],
                         toLeafLayer: bool = False) -> int:
        """
        Except when toLeafLayer is True, it is possible that some of the nodes in toLayer
        have no predecessors from fromLayer assigned to them at the end of this method.
        Such nodes might have to be explicitly removed after returning from here.
        """
        numNodes: int = 0
        leavesCovered: defaultdict = defaultdict(bool)
        sizeToLayer: int = len(toLayer)
        maxLeavesPerNode: int = min(2, sizeToLayer)
        for node in fromLayer:
            if toLeafLayer:
                # Select at most maxLeavesPerNode successors
                succPositions = sample(range(sizeToLayer), randint(1, maxLeavesPerNode))
            else:
                # Select successor positions randomly
                succPositions = sample(range(sizeToLayer), self._howManySuccessors())

            for succPos in succPositions:
                # Make connections between fromLayer and toLayer
                succNode: Node = toLayer[succPos]
                succNodeID: str
                if succNode is None:
                    succNodeID = Utils.generateID("node", TreeBuilder.numNodes)
                    succNode = Utils.newRandomNode(succNodeID,
                                                   self._supportedNodeTypes,
                                                   self._coreNodeClasses)
                    TreeBuilder.numNodes += 1
                    toLayer[succPos] = succNode
                    numNodes += 1
                node.addSuccessor(succNode)
                leavesCovered[succPos] = True
            # At the end of the loop, it is possible that some nodes in toLayer have no
            # predecessor nodes in fromLayer. In this case, the corresponding succPositions
            # in toLayer would continue to have None elements.

        if toLeafLayer:
            # If toLayer is the leaf layer, then we want to ensure that all leaf nodes have
            # at least one predecessor node.
            sizeFromLayer: int = len(fromLayer)
            for succPos in set(range(sizeToLayer)).difference(leavesCovered.keys()):
                randomFromPos = randint(0, sizeFromLayer - 1)
                fromLayer[randomFromPos].addSuccessor(toLayer[succPos])
                numNodes += 1
        return numNodes

    def generateNewGraph(self) -> Graph:
        """
        Returns:
          Graph: A tree-like graph, which mostly satisfies the specifications provided
                 in the constructor.
        """
        TreeBuilder.numGraphs += 1
        self._nodeLayers: List[List[Node]] = list()

        # 1. Create Graph ID.
        graphID: str = Utils.generateID("graph", TreeBuilder.numGraphs)
        graph: Graph = Graph(graphID, self._numRoots)

        # 2. Create desired no. of roots.
        rootLayer: List[Node] = list()
        for _ in range(self._numRoots):
            rNodeID = Utils.generateID("root", TreeBuilder.numNodes,)
            rNode = Utils.newRandomNode(rNodeID,
                                        self._supportedNodeTypes,
                                        self._coreNodeClasses)
            TreeBuilder.numNodes += 1
            graph.addRoot(rNode)
            rootLayer.append(rNode)
        self._nodeLayers.append(rootLayer)

        balNumInnerNodes: int = self._numInnerNodes
        # 3. For each currLayer, starting with roots
        currLayer: List[Node] = rootLayer
        nextLayer: List[Node] = [None] * self._numRoots * (self._branchingDegree + self._deltaRange)
        while(nextLayer is not None):
            balNumInnerNodes -= self._makeConnections(currLayer, nextLayer)
            try:
                # Since some of the nodes in nextLayer were not assigned any predecessors
                # from the currLayer, we simply remove those None values.
                while(True):
                    nextLayer.remove(None)
            except ValueError:
                # nextLayer is free from None nodes
                self._nodeLayers.append(nextLayer)
                currLayer = nextLayer

            if (0 >= balNumInnerNodes):
                nextLayer = None
            else:
                nextLayer = [None] * len(currLayer) * (self._branchingDegree + self._deltaRange)

        # 4. Create desired no. of leaves (nodes with `None` successors)
        leafLayer: List[Node] = list()
        for _ in range(self._numLeaves):
            lNodeID = Utils.generateID("leaf", TreeBuilder.numNodes)
            leafLayer.append(Utils.newRandomNode(lNodeID,
                                                 self._supportedNodeTypes,
                                                 self._coreNodeClasses))
            TreeBuilder.numNodes += 1
        _ = self._makeConnections(currLayer, leafLayer, True)
        self._nodeLayers.append(leafLayer)

        return graph


class DAGBuilder(TreeBuilder):
    """
    Directed Acyclic Graph Builder class
    """

    def __init__(self, **kwargs: str) -> None:
        """
        Construtor for the class takes exactly the same keyword arguments as TreeBuilder.
        Args:
          **kwargs(str): Must have the following keywords.
            KW_NUM_ROOTS - specifying the number of roots in any graph this builder creates.
            KW_NUM_LEAVES - specifying the number of leaves that any graph can have.
            KW_NUM_INNER_NODES - Besides roots and leaves, the approx. number of inner nodes.
            KW_BRANCHING_DEGREE - Approximate branching factor for each node.
            KW_SUPPORTED_NODE_TYPES - Concrete LoggerNode types that can be part of graphs.
        """
        super().__init__(**kwargs)
        # A graph with N layers will contain (N // _skipFraction) skip edges
        self._skipFraction: int = 3
        return

    def generateNewGraph(self) -> Graph:
        """
        This method first constructs a tree using the super class' generateNewGraph()
        method, and then adds some skip-level edges. Skip-level edges are edges that
        originate at a node on some level `i` and terminate at a node at some level
        `> (i + 1)`.

        Skip-level edges are only added to the graph if there are enough inner nodes to
        allow more than three levels.

        Returns:
          Graph: A directed acyclic graph, which mostly satisfies the specifications
                 provided in the constructor.
        """
        # 1. super().generateNewGraph()
        # 2. Randomly add skip edges, i.e. from layer_i to layer_>(i + 1)
        graph: Graph = super().generateNewGraph()

        numLevels: int = len(self._nodeLayers)
        if 4 > numLevels:
            return graph

        # Randomly choose levels from where some node will have an outgoing skip-edge
        fromLevels: List[int] = sample(range(numLevels - 2),
                                       numLevels // self._skipFraction)
        numNodesInLevel: int = 0
        for level in fromLevels:
            # For each level in the random list created above,
            # select a random node from which a skip-edge will
            # start
            numNodesInLevel = len(self._nodeLayers[level])
            fromNode: Node = self._nodeLayers[level][randint(0, numNodesInLevel - 1)]
            # Randomly select a lower level at which the skip-
            # edge will terminate
            toLevel: int = randint(level + 2, numLevels - 1)
            numNodesInLevel = len(self._nodeLayers[toLevel])
            # In the said lower level, randomly select a node
            # that will receive the skip-edge
            toNode: Node = self._nodeLayers[toLevel][randint(0, numNodesInLevel - 1)]
            fromNode.addSuccessor(toNode)
        return graph


class DCGBuilder(DAGBuilder):
    """
    Directed Cyclic Graph Builder class
    """

    # Additional keyword(s) for the kwargs expected by the constructor
    KW_REVERSE_EDGES = "REVERSE_EDGES"

    def __init__(self, **kwargs: str) -> None:
        """
        Construtor for the class takes all the keyword arguments of the super(-super) class
        TreeBuilder, and an additional one.
        Args:
          **kwargs(str): Must have the following keywords.
            KW_NUM_ROOTS - specifying the number of roots in any graph this builder creates.
            KW_NUM_LEAVES - specifying the number of leaves that any graph can have.
            KW_NUM_INNER_NODES - Besides roots and leaves, the approx. number of inner nodes.
            KW_BRANCHING_DEGREE - Approximate branching factor for each node.
            KW_SUPPORTED_NODE_TYPES - Concrete LoggerNode types that can be part of graphs.

            KW_REVERSE_EDGES - specifying whether or not there should be edges in the
                               "upward" direction, i.e. *to* level `i` *from* some level
                               `>i`. Reverse edges are only constructed if the graph has more
                               than 3 level.
        """
        super().__init__(**kwargs)

        self._reverseEdges: bool = False
        try:
            self._reverseEdges = bool(kwargs[DCGBuilder.KW_REVERSE_EDGES])
        except KeyError:
            self._reverseEdges = False
        # If a graph has N inner nodes, then (N // _loopFraction) nodes will
        # be set aside to appear in loops.
        self._loopFraction = 5
        self._numLoopNodes = ceil(self._numInnerNodes / self._loopFraction)
        self._numInnerNodes -= self._numLoopNodes
        return

    def _attachLoop(self, toNode: Node, ofSize: int) -> None:
        entityType: str = "loop_" + toNode.getID()
        currNode: Node = toNode
        loopNode: Node = None
        for _ in range(ofSize):
            lNodeID = Utils.generateID(entityType, DCGBuilder.numNodes)
            loopNode = Utils.newRandomNode(lNodeID,
                                           self._supportedNodeTypes,
                                           self._coreNodeClasses)
            DCGBuilder.numNodes += 1
            currNode.addSuccessor(loopNode)
            currNode = loopNode

        currNode.addSuccessor(toNode)
        return

    def generateNewGraph(self) -> Graph:
        """
        This method first constructs a DAG using the super class' generateNewGraph()
        method, and then adds loops and (optionally) reverse edges.

        Loops and reverse edges can only be added to a graph if there are enough
        inner nodes to allow more than three levels.

        Returns:
          Graph: A directed cyclic graph, which mostly satisfies the specifications
                 provided in the constructor.
        """
        # 1. super().generateNewGraph()
        graph: Graph = super().generateNewGraph()

        numLevels: int = len(self._nodeLayers)
        if 3 > numLevels:
            return graph

        # 2. Add loops
        numNodesPerLoop: int = 3  # FIXED FOR NOW
        numLoops: int = ceil(self._numLoopNodes / numNodesPerLoop)
        for _ in range(numLoops):
            level: int = randint(1, numLevels - 2)
            nodeNum: int = randint(0, len(self._nodeLayers[level]) - 1)
            toNode: Node = self._nodeLayers[level][nodeNum]
            self._attachLoop(toNode, numNodesPerLoop)

        if not self._reverseEdges:
            return graph

        # 3. Randomly add few reverse edges
        fromLevels: List[int] = sample(range(2, numLevels - 1),
                                       numLevels // self._skipFraction)
        for level in fromLevels:
            # For each level in the random list created above,
            # select a random node from which a reverse edge
            # will start
            numNodesInLevel = len(self._nodeLayers[level])
            fromNode: Node = self._nodeLayers[level][randint(0, numNodesInLevel - 1)]
            # Randomly select an upper level at which the
            # reverse edge will terminate
            toLevel: int = randint(0, level - 2)
            numNodesInLevel = len(self._nodeLayers[toLevel])
            # In the said upper level, randomly select a node
            # that will receive the skip-edge
            toNode: Node = self._nodeLayers[toLevel][randint(0, numNodesInLevel - 1)]
            fromNode.addSuccessor(toNode)

        return graph



import unittest
import os

from absynthe.cfg import Graph, UniformNode


class GraphTest(unittest.TestCase):

    def test_basicErrors(self):
        graph = Graph("EmptyGraph", 0)
        uselessNode = UniformNode("UselessNode")

        # Node cannot be added since 'graph' is created
        # with numRoots = 0
        with self.assertRaises(IndexError):
            graph.addRoot(uselessNode)
        return

    def test_graphSizeZero(self):
        graph = Graph("EmptyGraph", 0)
        self.assertEqual(graph.size(), 0)
        return

    def _buildDummyGraph(self) -> Graph:
        root1 = UniformNode("Root1")
        nodeR1S1 = UniformNode("Root1_Succ1")
        nodeR1S2 = UniformNode("Root1_Succ2")
        root1.addSuccessor(nodeR1S1)
        root1.addSuccessor(nodeR1S2)

        # Add successors to nodeR1S1
        nodeS1S1 = UniformNode("Succ1_Succ1")
        nodeS1S2 = UniformNode("Succ1_Succ2")
        nodeR1S1.addSuccessor(nodeS1S1)
        nodeR1S1.addSuccessor(nodeS1S2)

        root2 = UniformNode("Root2")
        nodeR2S1 = UniformNode("Root2_Succ1")
        nodeR2S2 = UniformNode("Root2_Succ2")
        root2.addSuccessor(nodeR2S1)
        root2.addSuccessor(nodeR2S2)

        # Create a cycle with edge nodeR1S1 -> root1
        nodeR1S1.addSuccessor(root1)

        graph = Graph("TestGraph", 2)
        graph.addRoot(root1)
        graph.addRoot(root2)

        return graph

    def test_graphSizeNonZero(self):
        testGraph = self._buildDummyGraph()
        self.assertEqual(testGraph.size(), 8)
        return

    def test_graphDump(self):
        filePath = os.path.dirname("../resources")+"/Graph_test_graphDump.gv"
        testGraph = self._buildDummyGraph()
        with open(filePath, "w") as ofp:
            testGraph.dumpDotFile(ofp)
        return


if __name__ == '__main__':
    unittest.main()



import unittest

from absynthe.graph_builder import TreeBuilder, DAGBuilder, DCGBuilder
from absynthe.cfg import Graph
import os


class TreeBuilderTest(unittest.TestCase):

    def test_treeTest(self):
        tree_kwargs = {TreeBuilder.KW_NUM_ROOTS: "2",
                       TreeBuilder.KW_NUM_LEAVES: "4",
                       TreeBuilder.KW_BRANCHING_DEGREE: "2",
                       TreeBuilder.KW_NUM_INNER_NODES: "16",
                       TreeBuilder.KW_SUPPORTED_NODE_TYPES: "SimpleLoggerNode"}

        testTreeBuilder = TreeBuilder(**tree_kwargs)
        testTree: Graph = testTreeBuilder.generateNewGraph()
        fileName = os.path.dirname(os.getcwd()
                                   + "/test/resources/") + "/TreeBuilder_test_TreeDump.gv"
        with open(fileName, 'w') as f:
            testTree.dumpDotFile(f)
        return

    def test_DAGTest(self):
        cfg_kwargs = {DAGBuilder.KW_NUM_ROOTS: "2",
                      DAGBuilder.KW_NUM_LEAVES: "4",
                      DAGBuilder.KW_BRANCHING_DEGREE: "3",
                      DAGBuilder.KW_NUM_INNER_NODES: "32",
                      DAGBuilder.KW_SUPPORTED_NODE_TYPES: "SimpleLoggerNode"}

        testDAGBuilder = DAGBuilder(**cfg_kwargs)
        testDAG: Graph = testDAGBuilder.generateNewGraph()
        fileName = os.path.dirname(os.getcwd()
                                   + "/test/resources/") + "/DAGBuilder_test_DAGDump.gv"
        with open(fileName, 'w') as f:
            testDAG.dumpDotFile(f)
        return

    def test_DCGTest(self):
        cfg_kwargs = {DCGBuilder.KW_NUM_ROOTS: "2",
                      DCGBuilder.KW_NUM_LEAVES: "4",
                      DCGBuilder.KW_BRANCHING_DEGREE: "3",
                      DCGBuilder.KW_NUM_INNER_NODES: "32",
                      DCGBuilder.KW_SUPPORTED_NODE_TYPES: "SimpleLoggerNode"}

        testDCGBuilder = DCGBuilder(**cfg_kwargs)
        testDCG: Graph = testDCGBuilder.generateNewGraph()
        fileName = os.path.dirname(os.getcwd()
                                   + "/test/resources/") + "/DCGBuilder_test_DCGDump.gv"
        with open(fileName, 'w') as f:
            testDCG.dumpDotFile(f)
        return


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import

from .node import Node, UniformNode, BinomialNode
from .logger_node import LoggerNode, SimpleLoggerNode
from .graph import Graph

__all__ = ["Node", "UniformNode", "BinomialNode",
           "LoggerNode", "SimpleLoggerNode",
           "Graph"]



from __future__ import print_function
from abc import ABC
from typing import List
from random import randint
from importlib import import_module

from absynthe.cfg import BinomialNode
from absynthe.cfg import LoggerNode


class Utils(ABC):
    """
    Utility methods for Nodes
    """

    # Node module
    LOGGER_NODE_MODULE = import_module("absynthe.cfg.logger_node")

    @staticmethod
    def generateID(entityType: str, entityNum: int) -> str:
        """
        Generates entity IDs from entity type and entity number.
        Args:
          entityType - Type of entity, e.g. "node", "leaf", "graph", etc.
          entityNum - Some number that uniquely identifies the entity.
        Returns:
          str - "entityType::entityNum"
        """
        return "::".join([entityType, str(entityNum)])

    @staticmethod
    def newRandomNode(nodeID: str,
                      loggerNodeOptions: List[str],
                      coreNodeOptions: List[str]) -> LoggerNode:
        """
        Instantiates an object of one of the concrete Node classes, wraps it
        inside one of the concrete LoggerNode classes, and returns the
        LoggerNode object.
        Args:
          nodeID - Identifier for the new node
          loggerNodeOptions - List of names of the possible, concrete
                              LoggerNode classes from which the return type
                              could be drawn
          coreNodeOptions - List of names of the possible, concrete Node
                            classes from which the core node of the above
                            LoggerNode could be drawn
        Returns:
          LoggerNode - A logger node
        """
        loggerNodeClassName = loggerNodeOptions[randint(0, len(loggerNodeOptions) - 1)]
        coreNodeClassName = coreNodeOptions[randint(0, len(coreNodeOptions) - 1)]
        logger_kwargs = {LoggerNode.KW_CORE_CLASS_NAME: coreNodeClassName,
                         LoggerNode.KW_IGNORE_PARAMS: "False"}

        if BinomialNode.__name__ == coreNodeClassName:
            logger_kwargs[BinomialNode.KW_P_VALUE] = "0.5"
        return getattr(Utils.LOGGER_NODE_MODULE, loggerNodeClassName)(nodeID, **logger_kwargs)



import unittest

from absynthe.cfg import UniformNode
from collections import defaultdict


class UniformNodeTest(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.UniformNode = UniformNode("Test_Node")
        succ1 = UniformNode("Successor_1")
        succ2 = UniformNode("Successor_2")
        succ3 = UniformNode("Successor_3")
        self.UniformNode.addSuccessor(succ1)
        self.UniformNode.addSuccessor(succ2)
        self.UniformNode.addSuccessor(succ3)
        # self.UniformNode.printDebugInfo(True)
        return

    def test_01_numSuccessors(self):
        self.assertEqual(self.UniformNode.getNumSuccessors(), 3)
        return

    def helper_diffWithinTolerance(self, val1: int, val2: int, tolerance: int = 150) -> bool:
        return (abs(val1 - val2) <= tolerance)

    def test_02_uniformDistribution(self):
        numSucc = self.UniformNode.getNumSuccessors()

        succCount = defaultdict(int)
        expectedIndividualCount = 5000
        for _ in range(numSucc * expectedIndividualCount):
            succName = self.UniformNode.getSuccessorAtRandom().getID()
            succCount[succName] += 1

        # all individual counts must be within 5% of expectedIndividualCount
        tolerance = (expectedIndividualCount * 5) // 100
        for i in range(numSucc):
            id = self.UniformNode.getSuccessorAt(i).getID()
            count = succCount[id]
            self.assertTrue(self.helper_diffWithinTolerance(count,
                                                            expectedIndividualCount,
                                                            tolerance))

    def test_03_deleteSuccessor(self):
        originalNumSuccessors = self.UniformNode.getNumSuccessors()
        expectedSucc3 = self.UniformNode.getSuccessorAt(-1)
        actualSucc3 = self.UniformNode.delLastSuccessor()
        newNumSuccessors = self.UniformNode.getNumSuccessors()

        self.assertEqual(expectedSucc3.getID(), actualSucc3.getID())
        self.assertEqual(originalNumSuccessors, newNumSuccessors + 1)

    def test_04_noSuccessors_pt1(self):
        testNode = UniformNode("testNode")
        self.assertEqual(testNode.getSuccessorAtRandom(), None)

    def test_04_noSuccessors_raises_IndexError(self):
        testNode = UniformNode("testNode")
        with self.assertRaises(IndexError):
            testNode.getSuccessorAt(0)


if __name__ == '__main__':
    unittest.main()



import unittest

from absynthe.cfg import Node


class NodeTest(unittest.TestCase):

    def test_raisesTypeError(self):
        with self.assertRaises(TypeError):
            _ = Node("AbstractClass? Really?")


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function

# Imports for LoggerNode
from .node import Node

from abc import abstractmethod
from sys import stderr
from typing import List
from random import randint
from importlib import import_module


class LoggerNode(Node):
    """
    An abstract wrapper class that provides the additional functionality of
    generating actual log messages. This class has a member field _coreNode
    that is a concrete subclass of Node, which ought to be available from
    importing absynthe.cfg.node.<ConcreteClassName>.

    This class is also a subclass of Node, and it implements all of the
    abstract methods of Node by delegating the call to _coreNode. This
    class additionally specifies methods to spit out actual, non-trivial,
    log messages.
    """

    # Keys expected in the kwargs parameter provided to the constructor
    # of this class.

    # To specify the class of the core node
    KW_CORE_CLASS_NAME = "CORE_CLASS_NAME"

    # Some nodes ignore flow params while spitting out the log message.
    KW_IGNORE_PARAMS = "IGNORE_PARAMS"

    # Type of messages supported
    MESG_TYPE_INFO = "INFO"
    MESG_TYPE_ERR = "ERROR"

    def __init__(self, id: str, **kwargs: str) -> None:
        """
        Constructor that internally creates an object of a concrete
        subclass of Node and assigns it to the member field _coreNode. The
        name of the concrete subclass is obtained through the keyword
        arguments, using the key 'coreClassName'.
        Args:
          id(str): Unique identifier for this node
          **kwargs(str): A dictionary of str -> str specifying, among
                         other things, the type of core node wrapped
                         inside this object. Must contain a key-value pair
                         against the key 'coreClassName' where the value is
                         the name of a concrete subclass of Node.

                         This argument is passed as is to the super class
                         constructor, as well as to the constructor of the
                         core node class. So it can contain keywords
                         specific to those constructors.
        Raises:
          AttributeError: If the value provided for 'coreClassName' is not a
                          valid class defined in absynthe.cfg.node module.
          TypeError: If the value provided for 'coreClassName' is a class
                     that cannot be instantiated, e.g. an abstract class.
        """
        super().__init__(id, **kwargs)
        coreID = id + "_Core"
        nodeModule = import_module("absynthe.cfg.node")

        # 1. Try to initialise the core node of this object
        self._coreNode: Node = None
        try:
            className = kwargs[LoggerNode.KW_CORE_CLASS_NAME]
        except KeyError as ke:
            # A core class name is mandatory...
            print(type(self).__name__, "ERROR: Keyword LoggerNode.KW_CORE_CLASS_NAME not provided",
                  className, file=stderr)
            raise ke
        try:
            self._coreNode = getattr(nodeModule, className)(coreID, **kwargs)
        except AttributeError as ae:
            # ...and shoud be a subclass of Node,
            print(type(self).__name__, "ERROR: Not a valid concrete subclass of Node - ",
                  className, file=stderr)
            raise ae
        except TypeError as te:
            # and should not be an abstract class.
            print(type(self).__name__, "ERROR: Class cannot be instantiated - ",
                  className, file=stderr)
            raise te

        # 2. Set other params used by methods that print out log messages
        self._ignoreParams = False
        try:
            ignoreParams = kwargs[LoggerNode.KW_IGNORE_PARAMS].lower()
            self._ignoreParams = (ignoreParams == "true")
        except KeyError:
            pass

        return

    def printDebugInfo(self, verbose: bool = False, printPrefix: str = "") -> None:
        """
        Override method to just print this node's ID ahead of the
        remainder of debug info that is generated from core ndoe.
        """
        print(printPrefix, self._id)
        self._coreNode.printDebugInfo(verbose, printPrefix)
        return

    def addSuccessor(self, successor: Node):
        """
        Override abstract method by delegating to core node.
        """
        self._coreNode.addSuccessor(successor)
        return

    def delLastSuccessor(self) -> Node:
        """
        Override abstract method by delegating to core node.
        """
        return self._coreNode.delLastSuccessor()

    def getNumSuccessors(self) -> int:
        """
        Override abstract method by delegating to core node.
        """
        return self._coreNode.getNumSuccessors()

    def getSuccessorAt(self, index: int) -> Node:
        """
        Override abstract method by delegating to core node.
        """
        return self._coreNode.getSuccessorAt(index)

    def getSuccessorAtRandom(self) -> Node:
        """
        Override abstract method by delegating to core node.
        """
        return self._coreNode.getSuccessorAtRandom()

    @abstractmethod
    def logInfo(self, timeStamp: str, params: List[str]) -> str:
        """
        Create a log message comprising of some fixed part and some variable part.
        The fixed part (i.e. the subseqeuence defining the log signature) ought to
        be immutable through the life of this object.

        The instantaneous parameter (e.g. the function call argument) is provided
        by the 'params' argument. This argument ought to be ignored if the field
        _ignoreParams is 'True'.
        Args:
          timeStamp(str): The time stamp associated with this log line
          params(List[str]): All the parameters that describe the variable part.
        Returns:
          str: A log message that respects _ignoreParams.
        """
        pass

    @abstractmethod
    def logError(self, timeStamp: str, params: List[str]) -> str:
        """
        Create an error message comprising of some fixed part and some variable part.
        The fixed part (i.e. the subseqeuence defining the log signature) ought to be
        immutable through the life of this object.

        The instantaneous parameter (e.g. the function call argument) is provided
        by the 'params' argument. This argument ought to be ignored if the field
        _ignoreParams is 'True'.
        Args:
          timeStamp(str): The time stamp associated with this log line
          params(List[str]): All the parameters that describe the variable part.
        Returns:
          str: A log message that respects _ignoreParams.
        """
        pass


class SimpleLoggerNode(LoggerNode):

    MIN_MESG_SIZE = 1
    MAX_MESG_SIZE = 4

    KW_PREFIX = "PREFIX"

    def __init__(self, id: str, **kwargs: str) -> None:
        super().__init__(id, **kwargs)
        meta: str = None
        try:
            meta = kwargs[SimpleLoggerNode.KW_PREFIX]
        except KeyError:
            pass
        except TypeError as te:
            print(type(self).__name__,
                  "ERROR: Expect keyword KW_PREFIX to provide `str` type.",
                  file=stderr)
            raise te
        prefix = meta if meta else ""
        self._fixedInfoMesg = self._createLoglineSignature(id, prefix)
        prefix += " ERROR"
        self._fixedErrMesg = self._createLoglineSignature(id, prefix)
        return

    def _createLoglineSignature(self, id: str, mesgPrefix: str = "") -> str:
        """
        Create and store a fixed signature that would be emitted with every log
        message. This node has different fixed signatures for info and error.
        """
        self._mesgLen = randint(self.MIN_MESG_SIZE, self.MAX_MESG_SIZE)
        fixedMesgList = [mesgPrefix]

        mesgInfixes = [self._id, self._coreNode.getID()]
        mesgInfix = mesgPrefix + "-" if not mesgPrefix == "" else None
        for i in range(self._mesgLen):
            fixedMesgList.append(" ")
            if mesgInfix:
                fixedMesgList.append(mesgInfix)
            for j in range(i + 1):
                fixedMesgList.append(mesgInfixes[j % 2])
                fixedMesgList.append("::")
            _ = fixedMesgList.pop()

        return "".join(fixedMesgList)

    def logInfo(self, timeStamp: str, params: List[str]) -> str:
        """
        Override abstract method of super class
        """
        return self._createMesg(timeStamp, params, LoggerNode.MESG_TYPE_INFO)

    def logError(self, timeStamp: str, params: List[str]) -> str:
        """
        Override abstract method of super class
        """
        return self._createMesg(timeStamp, params, LoggerNode.MESG_TYPE_ERR)

    def _createMesg(self, timeStamp: str,
                    params: List[str],
                    mesgType: str = LoggerNode.MESG_TYPE_INFO) -> str:
        mesgList = list()
        if mesgType == LoggerNode.MESG_TYPE_INFO:
            mesgList = [timeStamp, " ", self._fixedInfoMesg]
        elif mesgType == LoggerNode.MESG_TYPE_ERR:
            mesgList = [timeStamp, " ", self._fixedErrMesg]
        else:
            raise ValueError(type(self).__name__,
                             " ERROR: Unsupported type of log message -- ",
                             mesgType)

        if params is not None and not self._ignoreParams:
            mesgList.append(" [WITH_PARAMS] ")
            mesgList.extend(params)

        return "".join(mesgList)



import unittest
from scipy.stats import binom

from absynthe.cfg import BinomialNode
from collections import defaultdict


class BinomialNodeTest(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        node_kwargs = {BinomialNode.KW_P_VALUE: "0.5"}
        self.BinomialNode = BinomialNode("Test_Node", **node_kwargs)
        succ1 = BinomialNode("Successor_1", **node_kwargs)
        succ2 = BinomialNode("Successor_2", **node_kwargs)
        succ3 = BinomialNode("Successor_3", **node_kwargs)
        self.BinomialNode.addSuccessor(succ1)
        self.BinomialNode.addSuccessor(succ2)
        self.BinomialNode.addSuccessor(succ3)
        # self.BinomialNode.printDebugInfo(True)
        return

    def test_01_numSuccessors(self):
        self.assertEqual(self.BinomialNode.getNumSuccessors(), 3)
        return

    def helper_diffWithinTolerance(self, val1: int, val2: int, tolerance: int = 150) -> bool:
        return (abs(val1 - val2) <= tolerance)

    def test_02_binomialDistribution(self):
        numSucc = self.BinomialNode.getNumSuccessors()
        totalCount = numSucc * 5000

        succCount = defaultdict(int)
        expectedIndividualCounts = [totalCount * binom.pmf(k, numSucc - 1, self.BinomialNode._p) for k in range(numSucc)]
        for _ in range(totalCount):
            succName = self.BinomialNode.getSuccessorAtRandom().getID()
            succCount[succName] += 1

        # all individual counts must be within 5% of expectedIndividualCount
        tolerances = [(count * 5) // 100 for count in expectedIndividualCounts]
        for i in range(numSucc):
            id = self.BinomialNode.getSuccessorAt(i).getID()
            count = succCount[id]
            self.assertTrue(self.helper_diffWithinTolerance(count,
                                                            expectedIndividualCounts[i],
                                                            tolerances[i]))

    def test_03_deleteSuccessor(self):
        originalNumSuccessors = self.BinomialNode.getNumSuccessors()
        expectedSucc3 = self.BinomialNode.getSuccessorAt(-1)
        actualSucc3 = self.BinomialNode.delLastSuccessor()
        newNumSuccessors = self.BinomialNode.getNumSuccessors()

        self.assertEqual(expectedSucc3.getID(), actualSucc3.getID())
        self.assertEqual(originalNumSuccessors, newNumSuccessors + 1)

    def test_04_noSuccessors_pt1(self):
        node_kwargs = {BinomialNode.KW_P_VALUE: "0.5"}
        testNode = BinomialNode("testNode", **node_kwargs)
        self.assertEqual(testNode.getSuccessorAtRandom(), None)

    def test_04_noSuccessors_raises_IndexError(self):
        node_kwargs = {BinomialNode.KW_P_VALUE: "0.5"}
        testNode = BinomialNode("testNode", **node_kwargs)
        with self.assertRaises(IndexError):
            testNode.getSuccessorAt(0)


if __name__ == '__main__':
    unittest.main()



import imp
import os
from setuptools import setup, find_packages

version = imp.load_source(
    'absynthe.version', os.path.join('absynthe', 'version.py')).VERSION

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

setup(
    name = "absynthe",
    version = version,
    url = "https://github.com/chaturv3di/absynthe/",
    license = "Apache License 2.0",
    author = "Namit Chaturvedi",
    description = "A (branching) Behaviour Synthesizer",
    long_description = readme,
    long_description_content_type = "text/markdown",
    packages=find_packages(),
    classifiers = [
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: Log Analysis",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    python_requires = ">=3.6.0",
    install_requires = [
        'scipy==1.2.1',
        'numpy==1.16.4'],
    )



from __future__ import print_function
from sys import stderr

# Imports for Node
from abc import ABC, abstractmethod

# Imports for UniformNode
from random import randint

# Import for BinomialNode
from scipy.stats import binom
from random import random


class Node(ABC):
    """
    Abstract base class for nodes in a CFG. Each node would describe
    its own branching method based on the probability distribution
    governing the choice of which successor to return for traversal.
    """

    def __init__(self, id: str, **kwargs: str) -> None:
        """
        Args:
          id(str): An identifier for this node. It would help to
                   ensure that it is unique for a given use case.
          **kwargs(str): Keyword arguments specifying, among other
                   things, the probability mass function over the
                   choice of successors.
        """
        self._id = id
        self._successors = list()
        return

    # Concrete methods

    def getID(self) -> str:
        """
        Returns:
          str: The identifier for this node.
        """
        return self._id

    def getNumSuccessors(self) -> int:
        """
        Returns:
          int: Number of successors of this node.
        """
        return len(self._successors)

    def printDebugInfo(self, verbose: bool = False, printPrefix: str = "") -> None:
        """
        Prints information about this node. This must at least include
        the id and number of successors.
        Args:
          verbose(bool): If true, then the debug information ought to
                         contain debug information of the successors too.
        """
        print(printPrefix, self._id, "\b:", type(self).__name__)
        printPrefix += "\t"
        print(printPrefix, "Num Successors:", self.getNumSuccessors())

        for i in range(self.getNumSuccessors()):
            successor = self.getSuccessorAt(i)
            if verbose:
                successor.printDebugInfo(printPrefix=printPrefix)
            else:
                print(printPrefix, successor.getID())
        return

    # Abstract methods

    @abstractmethod
    def addSuccessor(self, successor: object) -> None:
        """
        Appends a successor Node to the list of successors. The
        order of successors is important since it governs the
        return value of delLastSuccessor() method.

        Additionally, it *might* govern the probability of selecting
        a successor if, e.g., the selection is made using a Gaussian
        distribution. In particular, the successor Nodes toward the
        centre of the list of successors would have the maximum
        probability of selection in that case.

        Args:
          successor(Node): A successor node to be added. `None` is a valid
            successor, since it denotes that this node could be a leaf node.
            Note that a node can have both `None` and non-`None` successors.
        """
        pass

    @abstractmethod
    def delLastSuccessor(self) -> object:
        """
        Delete the last successor that was added using the
        addSuccessor() method.
        """
        pass

    @abstractmethod
    def getSuccessorAt(self, index: int) -> object:
        """
        If the successors of this node are stored in an ordered list or,
        equivalently, uniquely mapped to 0-indexed integers, then this
        method should return the successor by the index value.

        Args:
          index(int): The index of a successor node.
        Returns:
          Node: The successor node at the given index.
        Raises:
          IndexError: If the index is out of bounds. Typically, index
                      takes values from 0 to getNumSuccessors - 1.
        """
        try:
            return self._successors[index]
        except IndexError as error:
            print("Illegal argument for index:", index, file=stderr)
            raise error

    @abstractmethod
    def getSuccessorAtRandom(self) -> object:
        """
        This method captures the probability distribution that governs
        the choice of successors while a behaviour is being synthesized.
        """
        pass


# ##
# Below are some concrete implementations of Node.
# ##

class UniformNode(Node):
    """
    A concrete implementation of Node that selects its successors
    *uniformly at random*.
    """

    def __init__(self, id, **kwargs) -> None:
        super().__init__(id, **kwargs)
        return

    def addSuccessor(self, successor: Node) -> None:
        """
        Appends a successor Node to the list of successors. Ensures that
        *at most one* successor is `None`.
        """
        if successor is None:
            try:
                # Check if None is already a successor
                _ = self._successors.index(None)
                return
            except ValueError:
                # If not, then add it.
                pass
        self._successors.append(successor)
        return

    def delLastSuccessor(self) -> Node:
        """
        Returns:
          Node: The last successor node that was added via
                addSuccessor() method.
        """
        return self._successors.pop()

    def getSuccessorAt(self, index: int) -> Node:
        return super().getSuccessorAt(index)

    def getSuccessorAtRandom(self) -> Node:
        if 0 == self.getNumSuccessors():
            return None

        randomIndex = randint(0, self.getNumSuccessors() - 1)
        return self._successors[randomIndex]


class BinomialNode(Node):
    """
    A concrete implementation of Node class that chooses its successors
    using a binomial PMF. For example, if a BinomialNode, with the parameter
    'p' = 0.5, has three successors then it will choose them with
    probabilities 0.25, 0.5, and 0.25 respectively.

    The second parameter 'n' of the binomial distribution is set
    automatically from the number of successors that are ultimately assigned
    to a BinomialNode.
    """

    # Keywords for kwargs expected by the constructor
    KW_P_VALUE = "P_VALUE"

    def __init__(self, id, **kwargs) -> None:
        """
        Construtor for the class.
        Args:
          **kwargs(str): Must have the following keywords.
            KW_P - Specifying the parameter 'p' of the binomial distribution
        """
        super().__init__(id, **kwargs)

        self._p = 0.5
        try:
            self._p = float(kwargs[BinomialNode.KW_P_VALUE])
        except KeyError as ke:
            print(type(self).__name__,
                  "ERROR - Probability value must be specified using the key",
                  "BinomialNode.KW_P_VALUE", file=stderr)
            raise ke
        return

    def addSuccessor(self, successor: Node) -> None:
        if successor is None:
            try:
                # Check if None is already a successor
                _ = self._successors.index(None)
            except ValueError:
                # Add it...
                pass
        self._successors.append(successor)
        return

    def delLastSuccessor(self) -> Node:
        return self._successors.pop()

    def getSuccessorAt(self, index: int) -> Node:
        return super().getSuccessorAt(index)

    def getSuccessorAtRandom(self) -> Node:
        """
        Returns:
          Node - A successor Node chosen according to the binomial
                 distribution parameterized by (n, p), where 'n' is the
                 number of successors of this node, and 'p' is provided in
                 the constructor. The probability of selecting a successor
                 at position k is pmf(k, n - 1, p), where 0 <= k <= n - 1.
        """
        numSuccessors: int = self.getNumSuccessors()

        if 0 == numSuccessors:
            return None

        p: float = random()

        for idx in range(numSuccessors):
            if p <= binom.cdf(idx, numSuccessors - 1, self._p):
                return self._successors[idx]
